%%%%%%%%%%%%
%
% Algorithm to calculate the core set of a set of points
% with probabilistic sampling (see JMLR, Tsang & Kwok, 2005
% for a core vector machine
%
%
% 
% IN:
%
% - fKPtr   - pointer to a kernel function e.g. fKPtr = @(X,Y)K(X,Y) - for a kernel matrix K
% - epsilon - eps of the minimum enclosing ball problem (MEB)
% - vDiags  - pre-calculated K(x,x) values for each x
%
% OUT:
% - S - set of core set points 
% - R - radius of the core set
% - vAlpha - loadings of the QP solution for S
%
% Note:
% 
% implemented by Frank-M. Schleif <fschleif@techfak.uni-bielefeld.de>
% 2011-2014
%
%% 
function [S,R,vAlpha] = core_set_one_two(fKPtr,epsilon,vDiags) % using only the kernel

% check input arguments
if(~exist('epsilon','var') || epsilon <= 0)% || epsilon > 1)
    epsilon = 0.1;   
end

% provide alternative quadratic-progr. solver if available
if(1 && exist('qpas','file')==3) % spm-qpc newcastle library available
    fPtrMinimalEnclosingBall = @(X,x0)MinimalEnclosingBallSPM(X,[]); % x0 is a dummy - warm start is not supported by spm-qpc
else
    fPtrMinimalEnclosingBall = @(X,x0)MinimalEnclosingBall(X,x0);
end

vDiags = vDiags(:);
N      = numel(vDiags);
iProbSampleSize = 59;
%% init 
fKDistPtr    = @(X,Y)(max(bsxfun(@minus,bsxfun(@plus,vDiags(X(:)),vDiags(Y(:))'),2*fKPtr(X,Y)'),0));
dTolerance   = (1+epsilon).^2; % for the epsilon problem
dOutlierPercentage = .001;  % 1% - 0.01       
SFull        = [1:N]; 
bSimple      = false;
if(bSimple)
    S            = randsample(N,2)';                                              % choose 2 random points (could be done more clever - but the cluster maybe a mix - so be careful)
else
    % identify raw center of the cloud
    vRandSampling= randsample(length(SFull),min(length(SFull),iProbSampleSize));               % some random samples    
    %Dists        = squareform(pdist(SFull(vRandSampling)',fKDistPtr));           % thats slower
    Dists        = fKDistPtr(SFull(vRandSampling),SFull(vRandSampling));
    dMeanDist    = mean(Dists(triu(Dists)~=0));
    vTypicalPairs= argmin(abs(Dists+diag(diag(ones(numel(vRandSampling))))*realmax-dMeanDist));
    [vCol,vRow]  = ind2sub(size(Dists),vTypicalPairs);
    S(1)         = vCol(1); % first pt (median of both should be somehow in the center)
    S(2)         = vRow(1); % snd pt 
end
SInit        = S;
SFull        = SFull(SFull~=S(1) & SFull~=S(2));                              % remove new added entry from SFull
vAlpha       = 0.5*ones(2,1);                                                 % default loadings in case of 2 points (initial center)
mKSS         = fKPtr(S,S);                                                    % kernel matrix
R            = (vAlpha' * diag(mKSS) -vAlpha' * mKSS * vAlpha);               % thats R^2 - initial radius
if(isempty(SFull))
    return;
end

%% core set calculation - inital step
iProbSampleSize = 59;%length(SFull);
vRandSampling= randsample(length(SFull),min(length(SFull),iProbSampleSize));

mK           = fKPtr(S,SFull(vRandSampling));                                                % kernel of core points to all out of radius points
mKSS         = fKPtr(S,S);      % SFull is without S    
vDistances   = (vDiags(SFull(vRandSampling))' -2 * vAlpha' * mK + vAlpha' * mKSS * vAlpha);   % get distances for MEB
%% loop  until MEB is found
iCounter     = 2;
while(~isempty(SFull))      
     vOutOfRadius     = find(vDistances > dTolerance * R);                    % identify points out of MEB
     if(length(vOutOfRadius)<=dOutlierPercentage*length(vDistances))          % if there are not to many it is still ok
         break;
     end     
     [~,I]            = max(vDistances(vOutOfRadius));                        % largest point away - original core-set heuristic
     iNewCoreIndex    = SFull(vRandSampling(vOutOfRadius(I)));                % identify new coreset point
     iCounter         = iCounter+1;
     S                = [S, iNewCoreIndex];                                   % add to core-set
     SFull            = SFull(SFull~=iNewCoreIndex);                          % remove new added entry from SFull and all which are in the circle
     mK               = fKPtr(S,S(end));
     mKSS             = [mKSS mK(1:end-1);mK'];
     %mKSS             = fKPtr(S,S);
     vAlpha0          = [vAlpha; 0.5]';
     vAlpha0          = vAlpha0./sum(vAlpha0);         
     [R,vAlpha]       = fPtrMinimalEnclosingBall(mKSS,vAlpha0);                                     % update radius (with warm start alpha if available     
     if(any(vAlpha>1+2*eps) || any(vAlpha+2*eps<0) || (sum(vAlpha)-1)>2*eps)
          warning('optimization failed - use last valid solution');
          break;
     end
     vRelevant           = find(vAlpha>1e-4);
     vAlpha              = vAlpha(vRelevant);
     vAlpha              = vAlpha./sum(vAlpha);
     S                   = S(vRelevant);
     mKSS                = mKSS(vRelevant,vRelevant);
     vRandSampling       = randsample(N-iCounter,min(N-iCounter,iProbSampleSize)); 
     if(~isempty(vRandSampling))
        mK               = fKPtr(S,SFull(vRandSampling));                                              % kernel of core points to all out of radius points
        vDistances       = max((vDiags(SFull(vRandSampling))' -2 * vAlpha' * mK + vAlpha' * mKSS * vAlpha),0); % calculate new distances
        iProbSampleSize  = min(59,numel(vRandSampling)); % if there are not enough points left
     end
end
% length(vAlpha)
% remove entries which vanished from the optimization - may happen 
if(sum(isnan(vAlpha))==0)
    S      = S(vAlpha>eps);
    vAlpha = vAlpha(vAlpha>eps);
    R      = sqrt(R); % return only R not R^2
else
    warning('kernel is invalid - solution may not be optimal - use last valid solution')
    vAlpha = vAlpha0;
    S      = S(vAlpha>eps);
    vAlpha = vAlpha(vAlpha>eps);
    R      = sqrt(R); % return only R not R^2    
end
if(numel(S)<2) % if the optimizer fails provide a backup
    S       = SInit;
    vAlpha  = 0.5*ones(2,1);                                                 % default loadings in case of 2 points (initial center)
    mKSS    = fKPtr(S,S);                                                    % kernel matrix
    R       = sqrt(vAlpha' * diag(mKSS) -vAlpha' * mKSS * vAlpha);           % thats R - initial radius
end

function [R,vAlpha] = MinimalEnclosingBall(mK,vAlpha0)
    N = size(mK,1);
    options = struct('Diagnostics','off','Algorithm','active-set','Display','off');
    vAlpha = quadprog(2*mK,[],[],[], ones(1,N),1, zeros(N,1),ones(N,1),vAlpha0,options); % here we use the dual formulation including a kernel (here linear, euclidean)
    R = (vAlpha' * diag(mK) -vAlpha' * mK * vAlpha);

function [R,vAlpha] = MinimalEnclosingBallSPM(mK,x0)
    N = size(mK,1);
    Z = zeros(N,1);
    vAlpha = qpas(2*mK,Z,[],[],ones(1,N),1,Z,ones(N,1)); % spm-qpc (newcastle) library
    R = (vAlpha' * diag(mK) -vAlpha' * mK * vAlpha);
    
    